@@ -2,7 +2,7 @@
 /* nm-openvpn-service - openvpn integration with NetworkManager
  *
  * Copyright (C) 2005 - 2008 Tim Niemueller <tim@niemueller.de>
- * Copyright (C) 2005 - 2008 Dan Williams <dcbw@redhat.com>
+ * Copyright (C) 2005 - 2010 Dan Williams <dcbw@redhat.com>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -54,6 +54,13 @@
 #include "nm-utils.h"
 #include "common/utils.h"
 
+#if !defined(DIST_VERSION)
+# define DIST_VERSION VERSION
+#endif
+
+static gboolean debug = FALSE;
+static GMainLoop *loop = NULL;
+
 #define NM_OPENVPN_HELPER_PATH		LIBEXECDIR"/nm-openvpn-service-openvpn-helper"
 
 G_DEFINE_TYPE (NMOpenvpnPlugin, nm_openvpn_plugin, NM_TYPE_VPN_PLUGIN)
@@ -64,6 +71,8 @@ typedef struct {
 	char *username;
 	char *password;
 	char *priv_key_pass;
+	char *proxy_username;
+	char *proxy_password;
 	GIOChannel *socket_channel;
 	guint socket_channel_eventid;
 } NMOpenvpnPluginIOData;
@@ -89,24 +98,29 @@ static ValidProperty valid_properties[] = {
 	{ NM_OPENVPN_KEY_CERT,                 G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CIPHER,               G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_COMP_LZO,             G_TYPE_BOOLEAN, 0, 0, FALSE },
-	{ NM_OPENVPN_KEY_MSSFIX,               G_TYPE_BOOLEAN, 0, 0, FALSE },
-	{ NM_OPENVPN_KEY_TUNNEL_MTU,           G_TYPE_INT, 0, G_MAXINT, FALSE },
-	{ NM_OPENVPN_KEY_FRAGMENT_SIZE,        G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_CONNECTION_TYPE,      G_TYPE_STRING, 0, 0, FALSE },
-	{ NM_OPENVPN_KEY_TAP_DEV,              G_TYPE_BOOLEAN, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_FRAGMENT_SIZE,        G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_KEY,                  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_LOCAL_IP,             G_TYPE_STRING, 0, 0, TRUE },
+	{ NM_OPENVPN_KEY_MSSFIX,               G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_PROTO_TCP,            G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_PORT,                 G_TYPE_INT, 1, 65535, FALSE },
+	{ NM_OPENVPN_KEY_PROXY_TYPE,           G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_PROXY_SERVER,         G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_PROXY_PORT,           G_TYPE_INT, 1, 65535, FALSE },
+	{ NM_OPENVPN_KEY_PROXY_RETRY,          G_TYPE_BOOLEAN, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_HTTP_PROXY_USERNAME,  G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_REMOTE,               G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_REMOTE_IP,            G_TYPE_STRING, 0, 0, TRUE },
+	{ NM_OPENVPN_KEY_RENEG_SECONDS,        G_TYPE_INT, 0, G_MAXINT, FALSE },
 	{ NM_OPENVPN_KEY_STATIC_KEY,           G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_STATIC_KEY_DIRECTION, G_TYPE_INT, 0, 1, FALSE },
 	{ NM_OPENVPN_KEY_TA,                   G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TA_DIR,               G_TYPE_INT, 0, 1, FALSE },
-	{ NM_OPENVPN_KEY_USERNAME,             G_TYPE_STRING, 0, 0, FALSE },
-	{ NM_OPENVPN_KEY_RENEG_SECONDS,        G_TYPE_INT, 0, G_MAXINT, FALSE },
+	{ NM_OPENVPN_KEY_TAP_DEV,              G_TYPE_BOOLEAN, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_TLS_REMOTE,	       G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_TUNNEL_MTU,           G_TYPE_INT, 0, G_MAXINT, FALSE },
+	{ NM_OPENVPN_KEY_USERNAME,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NULL,                                G_TYPE_NONE, FALSE }
 };
 
@@ -114,6 +128,7 @@ static ValidProperty valid_secrets[] = {
 	{ NM_OPENVPN_KEY_PASSWORD,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_CERTPASS,             G_TYPE_STRING, 0, 0, FALSE },
 	{ NM_OPENVPN_KEY_NOSECRET,             G_TYPE_STRING, 0, 0, FALSE },
+	{ NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD,  G_TYPE_STRING, 0, 0, FALSE },
 	{ NULL,                                G_TYPE_NONE, FALSE }
 };
 
@@ -277,8 +292,20 @@ nm_openvpn_disconnect_management_socket (NMOpenvpnPlugin *plugin)
 	g_io_channel_unref (io_data->socket_channel);
 
 	g_free (io_data->username);
+	g_free (io_data->proxy_username);
+
+	if (io_data->password)
+		memset (io_data->password, 0, strlen (io_data->password));
 	g_free (io_data->password);
 
+	if (io_data->priv_key_pass)
+		memset (io_data->priv_key_pass, 0, strlen (io_data->priv_key_pass));
+	g_free (io_data->priv_key_pass);
+
+	if (io_data->proxy_password)
+		memset (io_data->proxy_password, 0, strlen (io_data->proxy_password));
+	g_free (io_data->proxy_password);
+
 	g_free (priv->io_data);
 	priv->io_data = NULL;
 }
@@ -332,6 +359,33 @@ get_detail (const char *input, const char *prefix)
 	return ret;
 }
 
+static void
+write_user_pass (GIOChannel *channel,
+                 const char *authtype,
+                 const char *user,
+                 const char *pass)
+{
+	char *quser, *qpass, *buf;
+
+	/* Quote strings passed back to openvpn */
+	quser = ovpn_quote_string (user);
+	qpass = ovpn_quote_string (pass);
+	buf = g_strdup_printf ("username \"%s\" \"%s\"\n"
+	                       "password \"%s\" \"%s\"\n",
+	                       authtype, quser,
+	                       authtype, qpass);
+	memset (qpass, 0, strlen (qpass));
+	g_free (qpass);
+	g_free (quser);
+
+	/* Will always write everything in blocking channels (on success) */
+	g_io_channel_write_chars (channel, buf, strlen (buf), NULL, NULL);
+	g_io_channel_flush (channel, NULL);
+
+	memset (buf, 0, strlen (buf));
+	g_free (buf);
+}
+
 static gboolean
 handle_management_socket (NMVPNPlugin *plugin,
                           GIOChannel *source,
@@ -341,7 +395,6 @@ handle_management_socket (NMVPNPlugin *plugin,
 	NMOpenvpnPluginIOData *io_data = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin)->io_data;
 	gboolean again = TRUE;
 	char *str = NULL, *auth = NULL, *buf;
-	gsize written;
 
 	if (!(condition & G_IO_IN))
 		return TRUE;
@@ -355,25 +408,9 @@ handle_management_socket (NMVPNPlugin *plugin,
 	auth = get_detail (str, ">PASSWORD:Need '");
 	if (auth) {
 		if (strcmp (auth, "Auth") == 0) {
-			if (io_data->username != NULL && io_data->password != NULL) {
-				char *quser, *qpass;
-
-				/* Quote strings passed back to openvpn */
-				quser = ovpn_quote_string (io_data->username);
-				qpass = ovpn_quote_string (io_data->password);
-				buf = g_strdup_printf ("username \"%s\" \"%s\"\n"
-				                       "password \"%s\" \"%s\"\n",
-				                       auth, quser,
-				                       auth, qpass);
-				memset (qpass, 0, strlen (qpass));
-				g_free (qpass);
-				g_free (quser);
-
-				/* Will always write everything in blocking channels (on success) */
-				g_io_channel_write_chars (source, buf, strlen (buf), &written, NULL);
-				g_io_channel_flush (source, NULL);
-				g_free (buf);
-			} else
+			if (io_data->username != NULL && io_data->password != NULL)
+				write_user_pass (source, auth, io_data->username, io_data->password);
+			else
 				nm_warning ("Auth requested but one of username or password is missing");
 		} else if (!strcmp (auth, "Private Key")) {
 			if (io_data->priv_key_pass) {
@@ -386,11 +423,16 @@ handle_management_socket (NMVPNPlugin *plugin,
 				g_free (qpass);
 
 				/* Will always write everything in blocking channels (on success) */
-				g_io_channel_write_chars (source, buf, strlen (buf), &written, NULL);
+				g_io_channel_write_chars (source, buf, strlen (buf), NULL, NULL);
 				g_io_channel_flush (source, NULL);
 				g_free (buf);
 			} else
 				nm_warning ("Certificate password requested but private key password == NULL");
+		} else if (strcmp (auth, "HTTP Proxy") == 0) {
+			if (io_data->proxy_username != NULL && io_data->proxy_password != NULL)
+				write_user_pass (source, auth, io_data->proxy_username, io_data->proxy_password);
+			else
+				nm_warning ("HTTP Proxy auth requested but either proxy username or password is missing");
 		} else {
 			nm_warning ("No clue what to send for username/password request for '%s'", auth);
 			if (out_failure)
@@ -680,7 +722,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
                                  GError **error)
 {
 	NMOpenvpnPluginPrivate *priv = NM_OPENVPN_PLUGIN_GET_PRIVATE (plugin);
-	const char *openvpn_binary, *auth, *connection_type, *tmp;
+	const char *openvpn_binary, *auth, *connection_type, *tmp, *tmp2, *tmp3, *tmp4;
 	GPtrArray *args;
 	GSource *openvpn_watch;
 	GPid pid;
@@ -728,6 +770,34 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		add_openvpn_arg (args, tmp);
 	}
 
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_TYPE);
+	tmp2 = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_SERVER);
+	tmp3 = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_PORT);
+	tmp4 = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_PROXY_RETRY);
+	if (tmp && strlen (tmp) && tmp2 && strlen (tmp2)) {
+		if (!strcmp (tmp, "http")) {
+			add_openvpn_arg (args, "--http-proxy");
+			add_openvpn_arg (args, tmp2);
+			add_openvpn_arg (args, tmp3 ? tmp3 : "8080");
+			add_openvpn_arg (args, "'auto'");  /* Automatic proxy auth method detection */
+			if (tmp4)
+				add_openvpn_arg (args, "--http-proxy-retry");
+		} else if (!strcmp (tmp, "socks")) {
+			add_openvpn_arg (args, "--socks-proxy");
+			add_openvpn_arg (args, tmp2);
+			add_openvpn_arg (args, tmp3 ? tmp3 : "1080");
+			if (tmp4)
+				add_openvpn_arg (args, "--socks-proxy-retry");
+		} else {
+			g_set_error (error,
+				         NM_VPN_PLUGIN_ERROR,
+				         NM_VPN_PLUGIN_ERROR_BAD_ARGUMENTS,
+				         "Invalid proxy type '%s'.",
+				         tmp);
+			return FALSE;
+		}
+	}
+
 	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_COMP_LZO);
 	if (tmp && !strcmp (tmp, "yes"))
 		add_openvpn_arg (args, "--comp-lzo");
@@ -815,7 +885,7 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 		}
 	}
 
-	if (getenv ("OPENVPN_DEBUG")) {
+	if (debug) {
 		add_openvpn_arg (args, "--verb");
 		add_openvpn_arg (args, "10");
 	} else {
@@ -869,7 +939,10 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 
 	/* Up script, called when connection has been established or has been restarted */
 	add_openvpn_arg (args, "--up");
-	add_openvpn_arg (args, NM_OPENVPN_HELPER_PATH);
+	if (debug)
+		add_openvpn_arg (args, NM_OPENVPN_HELPER_PATH " --helper-debug");
+	else
+		add_openvpn_arg (args, NM_OPENVPN_HELPER_PATH);
 	add_openvpn_arg (args, "--up-restart");
 
 	/* Keep key and tun if restart is needed */
@@ -884,8 +957,9 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	/* Query on the management socket for user/pass */
 	add_openvpn_arg (args, "--management-query-passwords");
 
-	/* do not let openvpn setup routes, NM will handle it */
+	/* do not let openvpn setup routes or addresses, NM will handle it */
 	add_openvpn_arg (args, "--route-noexec");
+	add_openvpn_arg (args, "--ifconfig-noexec");
 
 	/* Now append configuration options which are dependent on the configuration type */
 	if (!strcmp (connection_type, NM_OPENVPN_CONTYPE_TLS)) {
@@ -979,24 +1053,28 @@ nm_openvpn_start_openvpn_binary (NMOpenvpnPlugin *plugin,
 	*/
 	if (   !strcmp (connection_type, NM_OPENVPN_CONTYPE_TLS)
 	    || !strcmp (connection_type, NM_OPENVPN_CONTYPE_PASSWORD)
-	    || !strcmp (connection_type, NM_OPENVPN_CONTYPE_PASSWORD_TLS)) {
-		NMOpenvpnPluginIOData *io_data;
+	    || !strcmp (connection_type, NM_OPENVPN_CONTYPE_PASSWORD_TLS)
+	    || nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME)) {
 
-		io_data = g_new0 (NMOpenvpnPluginIOData, 1);
+		priv->io_data = g_malloc0 (sizeof (NMOpenvpnPluginIOData));
 
 		tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_USERNAME);
-		io_data->username = tmp ? g_strdup (tmp) : NULL;
+		priv->io_data->username = tmp ? g_strdup (tmp) : NULL;
 		/* Use the default username if it wasn't overridden by the user */
-		if (!io_data->username && default_username)
-			io_data->username = g_strdup (default_username);
+		if (!priv->io_data->username && default_username)
+			priv->io_data->username = g_strdup (default_username);
 
 		tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_PASSWORD);
-		io_data->password = tmp ? g_strdup (tmp) : NULL;
+		priv->io_data->password = tmp ? g_strdup (tmp) : NULL;
 
 		tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_CERTPASS);
-		io_data->priv_key_pass = tmp ? g_strdup (tmp) : NULL;
+		priv->io_data->priv_key_pass = tmp ? g_strdup (tmp) : NULL;
 
-		priv->io_data = io_data;
+		tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME);
+		priv->io_data->proxy_username = tmp ? g_strdup (tmp) : NULL;
+
+		tmp = nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD);
+		priv->io_data->proxy_password = tmp ? g_strdup (tmp) : NULL;
 
 		nm_openvpn_schedule_connect_timer (plugin);
 	}
@@ -1040,6 +1118,11 @@ check_need_secrets (NMSettingVPN *s_vpn, gboolean *need_secrets)
 		/* Static key doesn't need passwords */
 	}
 
+	/* HTTP Proxy might require a password; assume so if there's an HTTP proxy username */
+	tmp = nm_setting_vpn_get_data_item (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_USERNAME);
+	if (tmp && !nm_setting_vpn_get_secret (s_vpn, NM_OPENVPN_KEY_HTTP_PROXY_PASSWORD))
+		*need_secrets = TRUE;
+
 	return ctype;
 }
 
@@ -1234,6 +1317,27 @@ nm_openvpn_plugin_new (void)
 	return plugin;
 }
 
+static void
+signal_handler (int signo)
+{
+	if (signo == SIGINT || signo == SIGTERM)
+		g_main_loop_quit (loop);
+}
+
+static void
+setup_signals (void)
+{
+	struct sigaction action;
+	sigset_t mask;
+
+	sigemptyset (&mask);
+	action.sa_handler = signal_handler;
+	action.sa_mask = mask;
+	action.sa_flags = 0;
+	sigaction (SIGTERM,  &action, NULL);
+	sigaction (SIGINT,  &action, NULL);
+}
+
 static void
 quit_mainloop (NMVPNPlugin *plugin, gpointer user_data)
 {
@@ -1244,10 +1348,36 @@ int
 main (int argc, char *argv[])
 {
 	NMOpenvpnPlugin *plugin;
-	GMainLoop *main_loop;
+	gboolean persist = FALSE;
+	GOptionContext *opt_ctx = NULL;
+
+	GOptionEntry options[] = {
+		{ "persist", 0, 0, G_OPTION_ARG_NONE, &persist, "Don't quit when VPN connection terminates", NULL },
+		{ "debug", 0, 0, G_OPTION_ARG_NONE, &debug, "Enable verbose debug logging (may expose passwords)", NULL },
+		{NULL}
+	};
 
 	g_type_init ();
 
+	/* Parse options */
+	opt_ctx = g_option_context_new ("");
+	g_option_context_set_translation_domain (opt_ctx, "UTF-8");
+	g_option_context_set_ignore_unknown_options (opt_ctx, FALSE);
+	g_option_context_set_help_enabled (opt_ctx, TRUE);
+	g_option_context_add_main_entries (opt_ctx, options, NULL);
+
+	g_option_context_set_summary (opt_ctx,
+		"nm-vpnc-service provides integrated OpenVPN capability to NetworkManager.");
+
+	g_option_context_parse (opt_ctx, &argc, &argv, NULL);
+	g_option_context_free (opt_ctx);
+
+	if (getenv ("OPENVPN_DEBUG"))
+		debug = TRUE;
+
+	if (debug)
+		g_message ("nm-openvpn-service (version " DIST_VERSION ") starting...");
+
 	if (system ("/sbin/modprobe tun") == -1)
 		exit (EXIT_FAILURE);
 
@@ -1255,15 +1385,15 @@ main (int argc, char *argv[])
 	if (!plugin)
 		exit (EXIT_FAILURE);
 
-	main_loop = g_main_loop_new (NULL, FALSE);
+	loop = g_main_loop_new (NULL, FALSE);
 
-	g_signal_connect (plugin, "quit",
-				   G_CALLBACK (quit_mainloop),
-				   main_loop);
+	if (!persist)
+		g_signal_connect (plugin, "quit", G_CALLBACK (quit_mainloop), loop);
 
-	g_main_loop_run (main_loop);
+	setup_signals ();
+	g_main_loop_run (loop);
 
-	g_main_loop_unref (main_loop);
+	g_main_loop_unref (loop);
 	g_object_unref (plugin);
 
 	exit (EXIT_SUCCESS);
